from musikla.core.events.event import DurationEvent
from .context import Context
from .voice import Voice
from .events import NoteEvent, MusicEvent
from .enumerable import merge_sorted
from typing import Any, Optional, List, Iterable, Tuple, Union
from itertools import islice
from fractions import Fraction

class MusicBuffer:
    def __init__ ( self ):
        self.buffer : List[MusicEvent] = []
    
    def append ( self, event : MusicEvent ):
        l = len( self.buffer )

        for i in range( l + 1 ):
            if i == l:
                self.buffer.append( event )
            elif self.buffer[ i ].timestamp > event.timestamp:
                self.buffer.insert( i, event )
                
                break

    def extend ( self, events : Iterable[MusicEvent] ):
        for event in events: self.append( event )

    def collect ( self, time : int = None ):
        if time == None:
            for event in self.buffer:
                yield event
            
            self.buffer.clear()
        else:
            while self.buffer and self.buffer[ 0 ].timestamp <= time:
                yield self.buffer.pop( 0 )
            
    def __len__ ( self ):
        return len( self.buffer )
    
    def __bool__ ( self ):
        return bool( self.buffer )

class Music:
    @classmethod
    def parallel ( cls, tracks ):
        return cls( merge_sorted( tracks, lambda note: note.timestamp ) )

    @classmethod
    def of ( cls, source : Union[MusicEvent, 'Music', Iterable[MusicEvent]] ) -> 'Music':
        if isinstance( source, Music ):
            return source
        elif isinstance( source, MusicEvent ):
            return cls( [ source ] )
        else:
            return cls( source )

    def __init__ ( self, notes = [] ):
        self.notes = notes
        self.shared_cache : Optional[SharedMusic] = None

    def __getitem__ ( self, key ) -> MusicEvent:
        return self.shared()[ key ]

    def len ( self, context : Context = None ) -> Union[Fraction, int]:
        range : Optional[Tuple[int, int]] = None
        
        for ev in self.shared():
            if range is None:
                range = ( ev.timestamp, ev.end_timestamp )
            else:
                range = ( range[ 0 ], max( range[ 1 ], ev.end_timestamp ) )

        if range is None:
            return Fraction(0)
            
        start, end = range

        if context is None:
            return end - start

        return context.voice.from_duration_absolute( end ) - context.voice.from_duration_absolute( start )

    def shared ( self ) -> 'SharedMusic':
        if self.shared_cache is None:
            self.shared_cache = SharedMusic( self )

        return self.shared_cache

    def expand ( self, context : Context ):
        for note in self:
            if isinstance( note, Music ):
                for subnote in note.expand( context ):
                    context.join( subnote.end_timestamp )

                    yield subnote
            else:
                context.join( note.end_timestamp )
                
                yield note

    def transform ( self, transformer : Any, *args, **kargs ) -> 'MusicGen':
        return MusicGen( self, lambda it: transformer.iter( it, *args, **kargs ) )

    def slice ( self, start = None, end = None, time : bool = True, cut : bool = False ) -> 'Music':
        if time:
            if cut:
                if start is not None and end is not None:
                    def _map ( event, index, start ):
                        if event.timestamp < start or event.end_timestamp > end:
                            timestamp = max( event.timestamp, start )

                            diff = timestamp - event.timestamp

                            return event.clone( timestamp = timestamp, duration = min( event.duration - diff, end - timestamp ) )

                        return event
                    
                    return self.filter( lambda e, i, s: e.end_timestamp > start and e.timestamp < end ).map( _map )
                elif start is not None:
                    def _map ( event, index, start ):
                        if event.timestamp < start:
                            diff = start - event.timestamp

                            return event.clone( timestamp = start, duration = event.duration - diff )

                        return event

                    return self.filter( lambda e, i, s: e.end_timestamp - s > start ).map( _map )
                elif end is not None:
                    def _map ( event, index, start ):
                        if event.end_timestamp > end:
                            return event.clone( duration = end )

                        return event

                    return self.filter( lambda e, i, s: e.timestamp < end ).map( _map )
            else:
                if start is not None and end is not None:
                    return self.filter( lambda e, i, s: e.timestamp >= start and e.end_timestamp <= end )
                elif start is not None:
                    return self.filter( lambda e, i, s: e.timestamp >= start )
                elif end is not None:
                    return self.filter( lambda e, i, s: e.end_timestamp <= end )
        else:
            if start is not None and end is not None:
                return MusicGen( self, lambda it: islice( it, start, end ) )
            elif start is not None:
                return MusicGen( self, lambda it: islice( it, start ) )
            elif end is not None:
                return MusicGen( self, lambda it: islice( it, 0, end ) )

        return self

    def first_note ( self, context : Context = None ) -> Optional[NoteEvent]:
        it = self.expand( context or Context() )

        try:
            value = next( it )

            while not isinstance( value, NoteEvent ):
                value = next( it )

            return value
        except StopIteration:
            return None
    
    def last_note ( self, context : Context ):
        note = None

        for value in self.expand( context ):
            if isinstance( value, NoteEvent ):
                note = value
                
        return note

    def map ( self, mapper ) -> 'MusicMap':
        return MusicMap( self, mapper )

    def filter ( self, predicate ) -> 'MusicFilter':
        return MusicFilter( self, predicate )

    def arp ( self, pattern : 'Music' = None ):
        from musikla.core.events.transformers import DecomposeChordsTransformer
        from musikla.core.theory import NotePitchClassesInv, NotePitchClassesIndexes

        chord = self.shared().transform( DecomposeChordsTransformer )
        pattern = pattern.shared()

        def _gen ( context : Context ):
            context = context or Context.default.fork( cursor = 0 )

            if pattern == None:
                baseline : Optional[int] = None
                
                for event in chord:
                    if baseline == None:
                        baseline = event.end_timestamp
                    else:
                        event = event.clone( timestamp = baseline )

                    context.join( event.end_timestamp )

                    yield event
            else:
                notes_pool : List[Any] = [ evt for evt in chord if isinstance( evt, NoteEvent ) ]

                len_pool = len( notes_pool )

                for event in pattern.transform( DecomposeChordsTransformer ).expand( context.fork() ):
                    if not isinstance( event, NoteEvent ):
                        context.join( event.end_timestamp )

                        yield event

                        continue

                    index = NotePitchClassesIndexes[ NotePitchClassesInv[ event.pitch_class ] ]
                    
                    if index >= 0 and index < len_pool:
                        archtype = notes_pool[ index ]

                        event = archtype.from_pattern( event )
            
                        context.join( event.end_timestamp )

                        yield event

        return MusicGen( None, lambda c: _gen( c ) )

    def repeat ( self, count : Union[int, bool] ):
        shared = self.shared()

        def _gen ( context ):
            if count == 0 or count == False: return
            
            if type(count) == int:
                for _ in range( 0, count ):
                    for event in shared.expand( context ):
                        yield event
            elif count == True:
                while True:
                    for event in shared.expand( context ):
                        yield event
        
        return MusicGen( None, lambda c: _gen(c) )

    def __pow__ ( self, length_or_music ):
        new_length : Union[Fraction, int] = Fraction( 0 )

        if type( length_or_music ) is float:
            new_length = Fraction( length_or_music )
        elif isinstance( length_or_music, Music ):
            new_length = length_or_music.len()
        elif isinstance( length_or_music, Fraction ):
            new_length = length_or_music
        else:
            return super().__pow__(length_or_music)

        music = self.shared()

        factor = new_length / music.len()

        def _stretch ( event : MusicEvent, index : int, start_time : int ):
            nonlocal factor

            timestamp = int( ( event.timestamp - start_time ) * factor ) + start_time

            if isinstance( event, DurationEvent ):
                return event.clone(
                    timestamp = timestamp,
                    value = Fraction( event.value * factor ).limit_denominator(32),
                    duration = event.voice.get_duration_absolute( event.value * factor )
                )
            else:
                return event.clone( timestamp = timestamp )

        return music.map( _stretch )

    def __mul__ ( self, other ):
        if type( other ) == int or type(other) is bool:
            return self.repeat( other )
        elif isinstance( other, Music ):
            return self.arp( other );
        else:
            # invalid
            ...

    def __add__ ( self, other ):
        return self.map( lambda ev, i, s: ev + other )
    
    def __sub__ ( self, other ):
        return self.map( lambda ev, i, s: ev - other )

    def __iter__ ( self ):
        if self.notes and hasattr( self.notes, '__iter__' ):
            for note in self.notes:
                yield note

class SharedMusic(Music):
    def __init__ ( self, base_music : Iterable ):
        self.base_music : Iterable = base_music
        self.shared_music : SharedIterator = SharedIterator( iter( base_music ) )

    def __getitem__ ( self, key ) -> MusicEvent:
        return self.shared_music.peek( key )

    def shared ( self ):
        return self

    def peek ( self ) -> Optional[MusicEvent]:
        event : Optional[MusicEvent] = None

        for ev in self:
            event = ev
            break
        
        return event
        
    def peek_many ( self, count : int ) -> List[MusicEvent]:
        events : List[MusicEvent] = []

        i = 0

        for ev in self:
            events.append( ev )
            
            i += 1

            if i >= count:
                break
        
        return events
    

    def retime ( self, context, offset, note ):
        if offset is None:
            offset = context.cursor - note.timestamp
        
        note = note.clone( timestamp = note.timestamp + offset )

        context.cursor = note.end_timestamp

        return offset, note

    def expand ( self, context : Context ):
        offset = None

        for note in self.shared_music:
            if isinstance( note, Music ):
                for subnote in note.expand( context ):
                    offset, subnote = self.retime( context, offset, subnote )

                    context.join( subnote.end_timestamp )

                    yield subnote
            else:
                offset, note = self.retime( context, offset, note )

                context.join( note.end_timestamp )

                yield note

    def __iter__ ( self ):
        for note in self.shared_music:
            yield note

class SharedIterator():
    def __init__ ( self, iterator ):
        self.iterator = iterator
        self.buffer : List = []
        self.stopped : bool = False

    def peek ( self, index : int ):
        start = len( self.buffer )

        if index < start:
            return self.buffer[ index ]

        if not self.stopped:
            for item in self.slice( start ):
                if start == index:
                    return item
                
                start += 1
        
        raise IndexError(index)

    def slice ( self, start : int = 0 ):
        i : int = start

        while not self.stopped or i < len( self.buffer ):
            if i < len( self.buffer ):
                i += 1

                yield self.buffer[ i - 1 ]
            else:
                try:
                    value = next( self.iterator )

                    self.buffer.append( value )

                    i += 1

                    yield value
                except StopIteration:
                    self.stopped = True

    def __iter__ ( self ):
        return self.slice()

    # def get_events ( self, context ):
    #     forked = self.context.fork( cursor = context.cursor )

    #     for event in self.node.eval( forked ):
    #         context.join( forked )
            
    #         yield event

    #     context.join( forked )
        
class TemplateMusic(Music):
    def __init__ ( self, notes = [] ):
        super().__init__( notes )
        self.shared_music : Optional[SharedMusic] = None

    def shared ( self ):
        return self
        
    def expand ( self, context : Context ):
        if self.shared_music == None:
            self.shared_music = SharedMusic( self.notes.eval( context.fork() ) )

        for note in self.shared_music.expand( context ):
            yield context.voice.revoice( note )

class MusicGen(Music):
    def __init__ ( self, base : Optional[Music], mapper ):
        super().__init__( [] )

        self.base : Optional[Music] = base
        self.mapper = mapper
        self.auto_expand : bool = base is not None

    def expand ( self, context : Context ):
        for event in self.mapper( self.base.expand( context ) if self.auto_expand else context ):
            yield event

    def __iter__ ( self ):
        for event in self.mapper( self.base if self.auto_expand else None ):
            yield event

class MusicMap(Music):
    def __init__ ( self, base : Music, mapper ):
        super().__init__( [] )

        self.base : Music = base
        self.mapper = mapper

    def expand ( self, context : Context ):
        start_time = 0
        index = 0

        for event in self.base.expand( context ):
            if index == 0:
                start_time = event.timestamp
            
            yield self.mapper( event, index, start_time )

            index += 1

    def __iter__ ( self ):
        start_time = 0
        index = 0

        for event in self.base:
            if index == 0:
                start_time = event.timestamp
            
            yield self.mapper( event, index, start_time )

            index += 1

class MusicFilter(Music):
    def __init__ ( self, base : Music, predicate ):
        super().__init__( [] )

        self.base : Music = base
        self.predicate = predicate

    def expand ( self, context : Context ):
        start_time = 0
        index = 0

        for event in self.base.expand( context ):
            if index == 0:
                start_time = event.timestamp
            
            if self.predicate( event, index, start_time ):
                yield event

            index += 1

    def __iter__ ( self ):
        start_time = 0
        index = 0

        for event in self.base:
            if index == 0:
                start_time = event.timestamp
            
            if self.predicate( event, index, start_time ):
                yield event

            index += 1



import sys
from typing import Any, List, Optional, Tuple, cast
from lark import Lark, Transformer, v_args
from musikla.core.theory import NoteAccidental, Note, Chord
from .abstract_syntax_tree import PythonNode
from .abstract_syntax_tree import NoteNode, ChordNode, MusicSequenceNode, MusicParallelNode, RestNode
from .abstract_syntax_tree.context_modifiers import LengthModifierNode, OctaveModifierNode, SignatureModifierNode, VelocityModifierNode, InstrumentModifierNode, TempoModifierNode, VoiceBlockModifier

from .abstract_syntax_tree.expressions import VariableExpressionNode, FunctionExpressionNode, FunctionChainExpressionNode, ListComprehensionNode
from .abstract_syntax_tree.expressions import StringLiteralNode, NumberLiteralNode, BoolLiteralNode, NoneLiteralNode
from .abstract_syntax_tree.expressions import ObjectLiteralNode, ArrayLiteralNode

from .abstract_syntax_tree.expressions import PlusBinaryOperatorNode, MinusBinaryOperatorNode, PowBinaryOperatorNode, MultBinaryOperatorNode, DivBinaryOperatorNode
from .abstract_syntax_tree.expressions import AndLogicOperatorNode, OrLogicOperatorNode

from .abstract_syntax_tree.expressions import GreaterComparisonOperatorNode, GreaterEqualComparisonOperatorNode
from .abstract_syntax_tree.expressions import EqualComparisonOperatorNode, NotEqualComparisonOperatorNode
from .abstract_syntax_tree.expressions import LesserComparisonOperatorNode, LesserEqualComparisonOperatorNode
from .abstract_syntax_tree.expressions import IsComparisonOperatorNode, IsNotComparisonOperatorNode
from .abstract_syntax_tree.expressions import InComparisonOperatorNode, NotInComparisonOperatorNode

from .abstract_syntax_tree.expressions import NotOperatorNode, GroupNode, BlockNode, PropertyAccessorNode

from .abstract_syntax_tree.statements import StatementsListNode, VariableDeclarationStatementNode, MultiVariableDeclarationStatementNode, FunctionDeclarationStatementNode
from .abstract_syntax_tree.statements import ForLoopStatementNode, WhileLoopStatementNode, IfStatementNode, ReturnStatementNode, ImportStatementNode

from .abstract_syntax_tree.macros import KeyboardDeclarationMacroNode, KeyboardShortcutMacroNode, KeyboardShortcutDynamicMacroNode, KeyboardShortcutComprehensionMacroNode, KeyboardForLoopMacroNode, KeyboardWhileLoopMacroNode, KeyboardIfMacroNode, KeyboardBlockMacroNode
from .abstract_syntax_tree.macros import VoiceDeclarationMacroNode

from fractions import Fraction

def isnumber ( n ):
    return type( n ) is int or type( n ) is float or isinstance( n, Fraction )

@v_args( tree = True )
class MusiklaTransformer(Transformer):
    def __init__(self, file : str = None, file_id : int = None, visit_tokens : bool = True):
        super().__init__(visit_tokens)

        self.file : Optional[str] = file
        self.file_id : Optional[int] = file_id

    def _get_position ( self, node, end_node = None ) -> Tuple[int, int, int]:
        end_pos = end_node.position[ 2 ] if end_node is not None and end_node.position is not None else node.meta.end_pos

        return ( self.file_id if self.file_id is not None else -1, node.meta.start_pos, end_pos )

    def start ( self, tree ):
        body = tree.children[ 0 ]

        if body is None:
            body = StatementsListNode([], position = self._get_position( tree ))

        if tree.children[ 1 ] is not None:
            if not isinstance( body, StatementsListNode ):
                body = StatementsListNode( [ body ], position = self._get_position( tree ) )

            body.statements.insert( 0, tree.children[ 1 ] )
        
        return body

    def body ( self, tree ):
        position = self._get_position( tree )

        children = list( tree.children )

        if any( ( c.hoisted for c in children ) ):
            children = [
                *filter( lambda c: c.hoisted, children ),
                *filter( lambda c: not c.hoisted, children )
            ]

        return StatementsListNode( children, position = position )


    def python ( self, tree ):
        position = self._get_position( tree )

        return PythonNode( tree.children[ 0 ], False, position = position )

    def assignment ( self, tree ):
        position = self._get_position( tree )

        operator = tree.children[ 1 ][ :-1 ]

        return VariableDeclarationStatementNode( tree.children[ 0 ], tree.children[ 2 ], operator = operator, position = position )

    def multi_assignment ( self, tree ):
        position = self._get_position( tree )

        operator = tree.children[ -2 ][ :-1 ]

        return MultiVariableDeclarationStatementNode( tree.children[ :-2 ], tree.children[ -1 ], operator = operator, position = position )

    def voice_assignment ( self, tree ):
        position = self._get_position( tree )

        name = tree.children[ 0 ]
        modifiers, parent = tree.children[ 1 ]
        
        return VoiceDeclarationMacroNode( 
            name, modifiers, parent,
            position = position
        )
    
    def voice_assignment_inherit ( self, tree ):
        return ( tree.children[ 1 ], VariableExpressionNode( tree.children[ 0 ] ) )
    
    def voice_assignment_base ( self, tree ):
        return ( tree.children[ 0 ], None )

    def import_global ( self, tree ):
        position = self._get_position( tree )

        return ImportStatementNode( tree.children[ 0 ], False, position = position )

    def import_local ( self, tree ):
        position = self._get_position( tree )

        return ImportStatementNode( tree.children[ 0 ], True, position = position )

    def for_variables ( self, tree ):
        return list( tree.children )

    def for_loop_head_range ( self, tree ):
        expression = FunctionExpressionNode( VariableExpressionNode( "range" ), [ tree.children[ 1 ], tree.children[ 2 ] ] )

        return ( tree.children[ 0 ], expression )

    def for_loop_head ( self, tree ):
        return ( tree.children[ 0 ], tree.children[ 1 ] )

    def for_loop_statement ( self, tree ):
        position = self._get_position( tree )

        variables, expression = tree.children[ 0 ]

        return ForLoopStatementNode( variables, expression, tree.children[ 1 ], position )

    def while_loop_statement ( self, tree ):
        position = self._get_position( tree )

        return WhileLoopStatementNode( tree.children[ 0 ], tree.children[ 1 ], position )

    def visit_while_loop_statement ( self, node, children ):
        position = self._get_position( node )

        return WhileLoopStatementNode( children.expression[ 0 ], children.body[ 0 ], position )
    
    def if_statement_else ( self, tree ):
        position = self._get_position( tree )

        return IfStatementNode( tree.children[ 0 ], tree.children[ 1 ], tree.children[ 2 ], position )

    def if_statement ( self, tree ):
        position = self._get_position( tree )

        return IfStatementNode( tree.children[ 0 ], tree.children[ 1 ], position = position )

    def if_body ( self, tree ):
        return tree.children[ 0 ]if tree.children else None

    def return_statement ( self, tree ):
        position = self._get_position( tree )

        if tree.children:
            return ReturnStatementNode( tree.children[ 0 ], position = position )

        return ReturnStatementNode( position = position )

    def parallel ( self, tree ):
        position = self._get_position( tree )

        return MusicParallelNode( list( tree.children ), position )

    def sequence ( self, tree ):
        position = self._get_position( tree )

        return MusicSequenceNode( list( tree.children ), position )

    def and_logic_op ( self, tree ):
        return AndLogicOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def or_logic_op ( self, tree ):
        return OrLogicOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def gte_comparison_op ( self, tree ):
        return GreaterEqualComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def gt_comparison_op ( self, tree ):
        return GreaterComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )
        
    def eq_comparison_op ( self, tree ):
        return EqualComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def neq_comparison_op ( self, tree ):
        return NotEqualComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def lte_comparison_op ( self, tree ):
        return LesserEqualComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def lt_comparison_op ( self, tree ):
        return LesserComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def isnot_comparison_op ( self, tree ):
        return IsNotComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )
        
    def is_comparison_op ( self, tree ):
        return IsComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def in_comparison_op ( self, tree ):
        return InComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def notin_comparison_op ( self, tree ):
        return NotInComparisonOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def sum_op ( self, tree ):
        return PlusBinaryOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def sub_op ( self, tree ):
        return MinusBinaryOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def pow_op ( self, tree ):
        return PowBinaryOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def mult_op ( self, tree ):
        return MultBinaryOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def div_op ( self, tree ):
        return DivBinaryOperatorNode( tree.children[ 0 ], tree.children[ 1 ], position = self._get_position( tree ) )

    def negation ( self, tree ):
        return NotOperatorNode( tree.children[ 0 ], position = self._get_position( tree ) )

    def accessor ( self, tree ):
        position = self._get_position( tree )

        expression = tree.children[ 0 ]
        accessor = tree.children[ 1 ]

        accessor.expression = expression
        accessor.position = position

        return accessor

    def method_call ( self, tree ):
        f, s, e = self._get_position( tree )

        property = self.property_accessor( tree )

        fn_call = tree.children[ 1 ]
        
        return FunctionChainExpressionNode( property, fn_call, position = ( f, s, e ) )

    def index_call ( self, tree ):
        f, s, e = self._get_position( tree )
        
        expression = tree.children[ 0 ]
        
        fn_call = tree.children[ 1 ]

        property = PropertyAccessorNode( cast( Any, None ), expression, False, position = (f, s, fn_call.position[ 2 ] - 1) )
        
        return FunctionChainExpressionNode( property, fn_call, position = ( f, s, e ) )

    def property_accessor ( self, tree ):
        s, f, _ = self._get_position( tree )
        
        identifier = StringLiteralNode( tree.children[ 0 ], position = (f, s, len( tree.children[ 0 ] ) ) )

        return PropertyAccessorNode( cast( Any, None ), identifier, True, position = identifier.position )

    def index_accessor ( self, tree ):
        position = self._get_position( tree )
        
        expression = tree.children[ 0 ]
        
        return PropertyAccessorNode( cast( Any, None ), expression, False, position = position )
        
    def python_expression ( self, tree ):
        position = self._get_position( tree )

        return PythonNode( tree.children[ 0 ], True, position = position )

    def function ( self, tree ):
        name = tree.children[ 0 ]

        fn_call = tree.children[ 1 ]

        _, s, _ = self._get_position( tree )
        f, _, e = fn_call.position

        fn_call.expression = VariableExpressionNode( name )
        fn_call.position = (f, s, e)

        return fn_call

    def function_call ( self, tree ):
        position = self._get_position( tree )
        
        positional, named = tree.children[ 0 ] or ([], {})
        
        return FunctionExpressionNode( None, positional, named, position )
        
    def function_call_chain ( self, tree ):
        _, s, _ = self._get_position( tree )

        right_hand = tree.children[ 1 ]
        f, rs, e = right_hand.position

        positional, named = tree.children[ 0 ] or ([], {})
        left_hand = FunctionExpressionNode( None, positional, named, (f, s, rs - 1) )
        
        return FunctionChainExpressionNode( left_hand, right_hand, (f, s, e) )

    def function_parameters ( self, tree ):
        positional = []
        named = {}

        for node in tree.children:
            if node.data == 'named_parameter':
                named[ node.children[ 0 ] ] = node.children[ 1 ]
            else:
                positional.append( node.children[ 0 ] )
        
        return positional, named

    def keyboard_declaration ( self, tree ):
        if len( tree.children ) == 3:
            return KeyboardDeclarationMacroNode( tree.children[ 2 ] or [], tree.children[ 0 ], tree.children[ 1 ] )

        return KeyboardDeclarationMacroNode( tree.children[ 1 ] or [], tree.children[ 0 ] )

    def keyboard_flags ( self, tree ):
        return list( tree.children )

    def keyboard_body ( self, tree ):
        return list( tree.children )

    def keyboard_for ( self, tree ):
        position = self._get_position( tree )
        
        var_name, expression = tree.children[ 0 ]

        return KeyboardForLoopMacroNode( var_name, expression, tree.children[ 1 ] or [], position = position )
    
    def keyboard_while ( self, tree ):
        position = self._get_position( tree )

        return KeyboardWhileLoopMacroNode( tree.children.expression, tree.children[ 1 ] or [], position = position )
    
    def keyboard_if ( self, tree ):
        position = self._get_position( tree )
        
        else_body = tree.children[ 2 ] if len( tree.children ) > 2 else None

        return KeyboardIfMacroNode( tree.children[ 0 ], tree.children[ 1 ] or [], else_body, position = position )
    
    def keyboard_block ( self, tree ):
        position = self._get_position( tree )
        
        return KeyboardBlockMacroNode( tree.children[ 0 ] if tree.children else None, position = position )

    def keyboard_shortcut ( self, tree ):
        position = self._get_position( tree )

        kind, key, flags = tree.children[ 0 ]

        expr = tree.children[ 2 ]

        args = tree.children[ 1 ]

        if kind == KeyboardShortcutComprehensionMacroNode:
            return KeyboardShortcutComprehensionMacroNode( key, flags, args, expr, position )
        elif kind == KeyboardShortcutDynamicMacroNode:
            return KeyboardShortcutDynamicMacroNode( key, flags, args, expr, position )
        else:
            return KeyboardShortcutMacroNode( key, args, expr, position )

    def keyboard_shortcut_key_static ( self, tree ):
        return (
            KeyboardShortcutMacroNode,
            list( tree.children ),
            None
        )

    def keyboard_shortcut_key_string ( self, tree ):
        return (
            KeyboardShortcutDynamicMacroNode,
            StringLiteralNode(tree.children[ 0 ]),
            list( tree.children[ 1: ] )
        )

    def keyboard_shortcut_key_dynamic ( self, tree ):
        return (
            KeyboardShortcutDynamicMacroNode,
            tree.children[ 0 ],
            list( tree.children[ 1: ] )
        )
    
    def keyboard_arguments ( self, tree ):
        if tree.children:
            return list( tree.children )
        
        return []

    def function_statements ( self, tree ):
        position = self._get_position( tree )

        name = tree.children[ 0 ]

        using = tree.children[ 2 ]

        body = tree.children[ 3 ]

        return FunctionDeclarationStatementNode( name, tree.children[ 1 ] or [], body, using, position )

    def function_expression ( self, tree ):
        position = self._get_position( tree )

        name = tree.children[ 0 ]
        
        using = tree.children[ 2 ]

        body = StatementsListNode( [ tree.children[ 3 ] ] )

        return FunctionDeclarationStatementNode( name, tree.children[ 1 ] or [], body, using, position )
    
    def using ( self, tree ):
        return list( tree.children )

    def arguments ( self, tree ):
        return list( tree.children )
    
    def argument_default ( self, tree ):
        return ( tree.children[ 1 ], tree.children[ 0 ], tree.children[ 2 ] )

    def argument ( self, tree ):
        return ( tree.children[ 1 ], tree.children[ 0 ], None )

    def argument_prefix ( self, tree ):
        return tree.children[ 0 ]

    def block ( self, tree ):
        position = self._get_position( tree )
        
        if not tree.children:
            return BlockNode( None, position = position )
        
        return BlockNode( tree.children[ 0 ], position = position )

    def block_call ( self, tree ):
        f, s, e = self._get_position( tree )
        
        if len( tree.children ) == 1:
            fn_call = tree.children[ 0 ]

            block = BlockNode( None )
        else:
            fn_call = tree.children[ 1 ]

            block = BlockNode( tree.children[ 0 ] )

        _, fs, fe = fn_call.position

        fn_call.expression = block
        fn_call.position = (f, s, fe)
        block.position = (f, s, fs - 1)

        return fn_call
        
    def group ( self, tree ):
        position = self._get_position( tree )
        
        return GroupNode( tree.children[ 0 ], position = position )

    def group_call ( self, tree ):
        f, s, e = self._get_position( tree )
        
        fn_call = tree.children[ 1 ]

        group = GroupNode( tree.children[ 0 ] )

        _, fs, fe = fn_call.position

        fn_call.expression = group
        fn_call.position = (f, s, fe)
        group.position = (f, s, fs - 1)

        return fn_call

    def string_literal ( self, tree ):
        position = self._get_position( tree )
        
        return StringLiteralNode( tree.children[ 0 ], position = position )
    
    def integer_literal ( self, tree ):
        position = self._get_position( tree )
        
        return NumberLiteralNode( tree.children[ 0 ], position = position )

    def float_literal ( self, tree ):
        position = self._get_position( tree )
        
        return NumberLiteralNode( tree.children[ 0 ], position = position )

    def bool_literal ( self, tree ):
        position = self._get_position( tree )
        
        return BoolLiteralNode( tree.children[ 0 ] == "true", position )

    def none_literal ( self, tree ):
        position = self._get_position( tree )

        return NoneLiteralNode( position )

    def array_literal ( self, tree ):
        position = self._get_position( tree )

        return ArrayLiteralNode( list( tree.children ), position = position )

    def object_literal ( self, tree ):
        position = self._get_position( tree )

        return ObjectLiteralNode( list( tree.children ), position = position )

    def object_item ( self, tree ):
        return ( tree.children[0], tree.children[1] )

    def object_key ( self, tree ):
        return tree.children[ 0 ]

    def chord_shortcut ( self, tree ):
        position = self._get_position( tree )
        
        value = Fraction( 1 )

        if len( tree.children ) == 3:
            value = tree.children[ 2 ]

        accidental, ( pitch_class, octave ) = tree.children[ 0 ]

        chord = tree.children[ 1 ]

        note = Note( 
            pitch_class = pitch_class,
            octave = octave,
            value = value,
            accidental = accidental
        )

        return ChordNode( Chord.from_abbreviature( note, chord, value ), position = position )

    def chord_manual ( self, tree ):
        position = self._get_position( tree )
        
        value = Fraction( 1 )

        x = len( tree.children )

        if tree.children and isnumber( tree.children[ -1 ] ):
            value = tree.children[ -1 ]
            x = -1

        notes : List[Note] = []

        for accidental, ( pitch_class, octave ) in tree.children[:x]:
            notes.append( Note( 
                pitch_class = pitch_class,
                octave = octave,
                value = value,
                accidental = accidental
            ) )

        return ChordNode( Chord( notes, None, value ), position = position )

    def CHORD_SUFFIX ( self, token ):
        return str( token )

    def rest ( self, tree ):
        position = self._get_position( tree )

        if tree.children:
            return RestNode( value = tree.children[ 0 ], visible = True, position = position )

        return RestNode( visible = True, position = position )

    def note ( self, tree ):
        position = self._get_position( tree )

        accidental, ( pitch_class, octave ) = tree.children[ 0 ]
        
        value = Fraction( 1 )

        if len( tree.children ) == 2:
            value = tree.children[ 1 ]
         
        note = Note( 
            pitch_class = pitch_class,
            octave = octave,
            value = value,
            accidental = accidental
        )

        return NoteNode( note, position )

    def note_value_int ( self, tree ):
        return tree.children[ 0 ]

    def note_value_frac ( self, tree ):
        if len( tree.children ) == 2:
            return Fraction( tree.children[ 0 ], tree.children[ 1 ] )
        else:
            return Fraction( 1, tree.children[ 0 ] )

    def note_pitch ( self, tree ):
        if len( tree.children ) == 1:
            return ( NoteAccidental.NONE, tree.children[ 0 ] )    

        return ( tree.children[ 0 ], tree.children[ 1 ] )

    def NOTE_ACCIDENTAL ( self, token ):
        if token == '^^':
            return NoteAccidental.DOUBLESHARP
        elif token == '^':
            return NoteAccidental.SHARP
        elif token == '_':
            return NoteAccidental.FLAT
        elif token == '__':
            return NoteAccidental.DOUBLEFLAT
        else:
            return NoteAccidental.NONE

    def NODE_PITCH_RAW ( self, token ):
        return Note.parse_pitch_octave( str( token ) )

    def tempo_mod ( self, tree ):
        return TempoModifierNode( tree.children[ 0 ], position = self._get_position( tree ) )

    def velocity_mod ( self, tree ):
        return VelocityModifierNode( tree.children[ 0 ], position = self._get_position( tree ) )

    def instrument_mod ( self, tree ):
        return InstrumentModifierNode( tree.children[ 0 ], position = self._get_position( tree ) )

    def length_mod ( self, tree ):
        return LengthModifierNode( tree.children[ 0 ], position = self._get_position( tree ) )

    def signature_mod ( self, tree ):
        position = self._get_position( tree )
        
        if len( tree.children ) == 2:
            return SignatureModifierNode( tree.children[ 0 ], tree.children[ 1 ], position )
        else:
            return SignatureModifierNode( lower = tree.children[ 1 ], position = position )

    def octave_mod ( self, tree ):
        return OctaveModifierNode( tree.children[ 0 ], position = self._get_position( tree ) )

    def voice_modifier ( self, tree ):
        position = self._get_position( tree )

        return VoiceBlockModifier( tree.children[ 1 ], tree.children[ 0 ], position = position )

    def VOICE_CALL ( self, token ):
        return token[1:-1]

    def VOICE_IDENTIFIER ( self, token ):
        return token[1:]

    def variable_call ( self, tree ):
        fn_call = tree.children[ 1 ]

        fn_call.position = self._get_position( tree, fn_call )

        fn_call.expression = VariableExpressionNode( tree.children[ 0 ] )

        return fn_call

    def variable_name ( self, tree ):
        position = self._get_position( tree )

        return VariableExpressionNode( tree.children[ 0 ], position = position )

    def FUNCTION_CALL ( self, token ):
        return token[:-1]

    def VARIABLE_CALL ( self, token ):
        return token[1:-1]

    def NAMED_IDENTIFIER ( self, token ):
        return token.strip( " \t\r\n=" )

    def VARIABLE_NAME ( self, token ):
        return token[1:]

    def IDENTIFIER ( self, token ):
        return str( token )

    def ALPHANUMERIC ( self, token ):
        return str( token )

    def INTEGER ( self, token ):
        return int( token )

    def FLOAT ( self, token ):
        return float( token )

    def STRING ( self, token ):
        return StringLiteralNode.escaped( token[1:-1], quote = token[ 0 ] )



from typing import List, Optional

class CodePrinter:
    def __init__ ( self, ident : int = 4 ):
        self.ident : int = ident
        self.stack : List['CodeBlock'] = [ CodeBlock( 0, None, None, 'always' ) ]

    def block ( self, opening : Optional[str] = '{', closing : Optional[str] = '}', multiline : str = 'auto' ) -> 'CodeBlockContext':
        return CodeBlockContext( self, opening, closing, multiline )

    def begin_block ( self, opening : Optional[str] = '{', closing : Optional[str] = '}', multiline : str = 'auto' ):
        if opening is not None:
            self.add_token( opening )

        self.stack.append( CodeBlock( self.ident * len( self.stack ), opening, closing, multiline ) )

    def begin_line ( self ):
        self.stack[ -1 ].add_line()

    def add_token ( self, token : str ):
        self.stack[ -1 ].add_token( token )

    def end_block ( self ):
        block = self.stack.pop()

        self.stack[ -1 ].add_block( block )

    def clear ( self ):
        self.stack = [ CodeBlock( 0, None, None, 'always' ) ]

    def print ( self, node ) -> str:
        node.to_source( self )

        result = ''.join( self.stack[ -1 ].tokens )

        self.clear()

        return result

class CodeBlockContext:
    def __init__ ( self, printer : CodePrinter, opening : Optional[str] = '{', closing : Optional[str] = '}', multiline_strat : str = 'auto' ):
        self.printer : CodePrinter = printer
        self.opening : Optional[str] = opening
        self.closing : Optional[str] = closing
        self.multiline_strat : str = multiline_strat
    
    def __enter__ ( self ):
        self.printer.begin_block( self.opening, self.closing, self.multiline_strat )

        return self
    
    def __exit__ ( self, exc_type, exc_val, exc_tb ):
        self.printer.end_block()

class CodeBlock:
    def __init__ ( self, ident : int, opening : Optional[str] = '{', closing : Optional[str] = '}', multiline_strat : str = 'auto' ):
        self.ident : int = ident
        self.opening : Optional[str] = opening
        self.closing : Optional[str] = closing
        self.multiline_strat : str = multiline_strat
        self.tokens : List[str] = []
        self.line_number : int = 0

    @property
    def multiline ( self ) -> bool:
        return ( self.line_number > 1 and self.multiline_strat != 'never' )\
            or ( self.multiline_strat == 'always' and self.line_number == 1 )

    def get_new_line ( self ) -> str:
        linebreak = ( '\n' if self.line_number > 0 or self.opening is not None else '' )

        return linebreak + ( ' ' * self.ident )

    def add_line ( self ):
        if self.line_number == 1 and self.multiline_strat == 'auto':
            self.tokens.insert( 0, self.get_new_line() )

        if ( self.line_number > 0 and self.multiline_strat != 'never' ) or self.multiline_strat == 'always':
            self.tokens.append( self.get_new_line() )
        
        self.line_number += 1

    def add_token ( self, token : str ):
        if self.line_number == 0:
            self.add_line()

        self.tokens.append( token )

    def add_block ( self, block : 'CodeBlock' ):
        self.tokens.extend( block.tokens )

        if block.multiline:
            self.add_line()

        if block.closing is not None:
            self.add_token( block.closing )


from musikla.parser.printer import CodePrinter
from typing import List, Tuple
from musikla.core import Value
from ..node import Node
from ..music_node import MusicSequenceBase

class ForLoopStatementNode( MusicSequenceBase ):
    def __init__ ( self, variable : List[str], it : Node, body : Node, position : Tuple[int, int, int] = None ):
        from ..expressions.block_node import BlockNode

        super().__init__( position )

        self.variables : List[str] = variable
        self.it : Node = it
        self.body : Node = body
        
        if isinstance( self.body, BlockNode ):
            self.body.fork_context = False
            self.body.create_stack_frame = False

    def values ( self, context ):
        for i in Value.eval( context, self.it ):
            forked = context.fork( symbols = context.symbols.fork( opaque = False ) )

            if len( self.variables ) == 1:
                forked.symbols.assign( self.variables[ 0 ], i, local = True )
            else:
                for k, name in enumerate( self.variables ):
                    forked.symbols.assign( name, i[ k ], local = True )

            yield Value.eval( forked, self.body )

            context.join( forked )

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'for ' )

        with printer.block( '(', ')' ):
            printer.add_token( ", ".join( [ "$" + name for name in self.variables ] ) + " in " )
            
            self.it.to_source( printer )

        self.body.to_source( printer )




from .event import DurationEvent, VoiceEvent
from ..voice import Voice
from ..theory import Note, NoteAccidental, Interval
from typing import Optional, cast

class NoteOnEvent ( VoiceEvent ):
    @staticmethod
    def from_pitch ( timestamp = 0, pitch = 0, voice : Voice = None, velocity = 127, tied : bool = False, staff : Optional[int] = 0 ):
        note = Note.from_pitch( pitch )
        
        return NoteOnEvent(
            timestamp = timestamp,
            velocity = velocity,
            voice = voice,
            pitch_class = note.pitch_class, 
            octave = note.octave, 
            accidental = note.accidental,
            tied = tied,
            staff = staff
        )

    def __init__ ( self, timestamp = 0, pitch_class = 0, octave = 4, accidental = NoteAccidental.NONE, velocity = 0, voice : Voice = None, tied : bool = False, staff : Optional[int] = 0 ):
        self.pitch_class : int = pitch_class
        self.octave : int = octave
        self.accidental : int = accidental
        self.velocity : int = velocity
        self.tied : bool = tied

        self._disabled : bool = False

        super().__init__( timestamp, voice, staff )

    @property
    def note ( self ) -> Note:
        return Note( self.pitch_class, self.octave, self.accidental )

    def note_off ( self, timestamp : int ) -> 'NoteOffEvent':
        return NoteOffEvent( timestamp, self.pitch_class, self.octave, self.accidental, self.voice, self.tied, self.staff )

    def __int__ ( self ):
        return int( self.note )

    def __str__ ( self ):
        return str( self.note ) + "(On)"


class NoteOffEvent ( VoiceEvent ):
    @staticmethod
    def from_pitch ( timestamp = 0, pitch = 0, voice : Voice = None, tied : bool = False, staff : Optional[int] = 0 ):
        note = Note.from_pitch( pitch )
        
        return NoteOffEvent(
            timestamp = timestamp,
            voice = voice,
            pitch_class = note.pitch_class, 
            octave = note.octave, 
            accidental = note.accidental,
            tied = tied,
            staff = staff
        )

    def __init__ ( self, timestamp = 0, pitch_class = 0, octave = 4, accidental = NoteAccidental.NONE, voice : Voice = None, tied : bool = False, staff : Optional[int] = 0 ):
        self.pitch_class : int = pitch_class
        self.octave : int = octave
        self.accidental : int = accidental
        self.tied : bool = tied

        super().__init__( timestamp, voice, staff )

    @property
    def note ( self ) -> Note:
        return Note( self.pitch_class, self.octave, self.accidental )

    def __int__ ( self ):
        return int( self.note )

    def __str__ ( self ):
        return str( self.note ) + "(Off)"

class NoteEvent( DurationEvent ):
    @staticmethod
    def from_pitch ( timestamp = 0, pitch = 0, duration = 4, voice : Voice = None, velocity = 127, value = None, tied : bool = False, staff : Optional[int] = 0 ):
        note = Note.from_pitch( pitch )
        
        return NoteEvent(
            timestamp = timestamp,
            duration = duration,
            value = value,
            velocity = velocity,
            voice = voice,
            pitch_class = note.pitch_class, 
            octave = note.octave, 
            accidental = note.accidental,
            tied = tied,
            staff = staff
        )

    def __init__ ( self, timestamp = 0, pitch_class = 0, duration = 4, octave = 4, voice : Voice = None, velocity = 127, accidental = NoteAccidental.NONE, value = None, tied : bool = False, staff : Optional[int] = 0 ):
        super().__init__( timestamp, duration, value, voice, staff )

        self.pitch_class = pitch_class
        self.octave = octave
        self.velocity = velocity
        self.accidental = accidental
        self.tied = tied

    def music ( self ):
        from ..music import SharedMusic

        return SharedMusic( [ self ] )

    @property
    def note ( self ) -> Note:
        return Note( self.pitch_class, self.octave, self.accidental, self.value )

    @property
    def note_on ( self ) -> NoteOnEvent:
        return NoteOnEvent( self.timestamp, self.pitch_class, self.octave, self.accidental, self.velocity, self.voice, self.tied, self.staff )

    @property
    def note_off ( self ) -> NoteOffEvent:
        return NoteOffEvent( self.timestamp + self.duration, self.pitch_class, self.octave, self.accidental, self.voice, self.tied, self.staff )

    def from_pattern ( self, pattern : 'NoteEvent' ) -> 'NoteEvent':
        return cast( NoteEvent, self.clone( 
             timestamp = pattern.timestamp,
             octave = self.octave + ( pattern.octave - pattern.voice.octave ) + 1,
             value = pattern.value,
             duration = pattern.duration,
             velocity = pattern.velocity,
             voice = pattern.voice,
             staff = pattern.staff
        ) )

    def with_pitch ( self, pitch : int ) -> 'NoteEvent':
        return NoteEvent.from_pitch(
            timestamp = self.timestamp,
            pitch = pitch,
            duration = self.duration, 
            voice = self.voice, 
            velocity = self.velocity, 
            value = self.value,
            tied = self.tied,
            staff = self.staff
        )

    def __lt__ ( self, other ):
        if other is None: return False

        return int( self ) < int( other )

    def __le__ ( self, other ):
        if other is None: return False

        return int( self ) <= int( other )

    def __eq__ ( self, other ):
        if other is None or not isinstance( other, NoteEvent ):
            return False

        return int( self ) == int( other )

    def __add__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitch( int( self ) + interval )
        elif isinstance( interval, Interval ):
            return self.with_pitch( int( self ) + int( interval ) )
        else:
            return self

    def __sub__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitch( int( self ) - interval )
        elif isinstance( interval, Interval ):
            return self.with_pitch( int( self ) - int( interval ) )
        else:
            return self

    def __int__ ( self ):
        return int( self.note )

    def __repr__ ( self ):
        return f'[{self.timestamp}]' + str( self.note )

    def __str__ ( self ):
        return str( self.note )




from musikla.parser.printer import CodePrinter
from typing import Tuple
from .expression_node import ExpressionNode

class StringLiteralNode( ExpressionNode ):
    def __init__ ( self, value, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.value = value

    def __eval__ ( self, context, assignment : bool = False ):
        return self.value

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( "'" + self.value + "'" )

    @staticmethod
    def escaped ( value, quote : str = '"' ) -> 'StringLiteralNode':
        escaped_str = None

        l = len( value )
        v = None
        rv = None
        skip = False
        cursor = 0

        for i in range( l ):
            if skip:
                skip = False

                continue

            if value[ i ] == "\\":
                if l > i + 1:
                    v = value[ i + 1 ]
                    rv = None

                    if v == "\\":
                        rv = "\\"
                    elif v == "n":
                        rv = "\n"
                    elif v == "r":
                        rv = "\r"
                    elif v == "t":
                        rv = "\t"
                    elif v == quote:
                        rv = quote
                    
                    if rv is not None:
                        skip = True

                        if escaped_str is None:
                            escaped_str = value[ :i ] + rv
                        else:
                            escaped_str += value[ cursor:i ] + rv
                        
                        cursor = i + 2

        if escaped_str is None:
            escaped_str = value
        else:
            escaped_str += value[cursor:]
        
        return escaped_str



from .. import MusicNode

class ContextModifierNode( MusicNode ):
    def apply ( self, voice ):
        pass

    def modify ( self, context ):
        pass

    def get_events ( self, context ):
        return self.modify( context )



import time
import hashlib
import tempfile
from pathlib import Path
from lark import Lark, UnexpectedInput
from typing import Optional
from .error_reporter import ErrorReporter
from .abstract_syntax_tree import Node
from .transformer import MusiklaTransformer


class Parser():
    def __init__ ( self, read_cache : bool = True, write_cache : bool = True ):
        self.debug : bool = False

        self.time_spent : float = 0
        self.lark_time_spent : float = 0
        self.read_cache : bool = read_cache
        self.write_cache : bool = write_cache

        self.internal_parser = self.create_parser( 
            read_cache = self.read_cache,
            write_cache = self.write_cache
        )

        self.specific_parsers = dict()

    def get_cache_path ( self, rule : Optional[str], content : str ) -> Path:
        cache_path = Path( tempfile.gettempdir() )

        if rule is not None:
            content = rule + "|" + content
        
        dig = hashlib.md5( content.encode() ).hexdigest()
            
        return cache_path / ( 'grammar.lark.pickle.' + dig + '.tmp' )

    def create_parser ( self, rule : str = None, read_cache : bool = True, write_cache : bool = True ):
        with open( Path( __file__ ).parent / "grammar.lark", "r" ) as f:
            grammar_content = f.read()

            cache_path = self.get_cache_path( rule, grammar_content ) if read_cache or write_cache else None

            parser = None

            if read_cache and cache_path.exists():
                with open( cache_path, "rb" ) as f:
                    parser = Lark.load( f )
            else:
                parser = Lark( grammar_content, start = rule or 'start', 
                    parser='lalr', debug=False, propagate_positions = True, 
                    maybe_placeholders = True )
            
                if write_cache:
                    with open( cache_path, "wb" ) as f:
                        parser.save( f )

            return parser

    def get_parser ( self, rule : str = None ):
        if rule is None or rule == "start":
            return self.internal_parser
        else:
            if rule not in self.specific_parsers:
                self.specific_parsers[ rule ] = self.create_parser( rule )
            
            return self.specific_parsers[ rule ]

    def parse ( self, expression, file : str = None, file_id : int = None, rule : str = None ) -> Node:
        try:
            parser = self.get_parser( rule )

            start_time = time.time()
            tree = MusiklaTransformer( file = file, file_id = file_id ).transform( parser.parse( expression ) )
            self.time_spent += time.time() - start_time

            return tree;
        except UnexpectedInput as err:
            raise ParserError( err, expression, file )

    def parse_file ( self, file ) -> Node:
        with open( file, 'r', encoding="utf-8" ) as f:
            return self.parse( f.read(), file = file )

class ParserError(Exception):
    def __init__ ( self, err : UnexpectedInput, contents : str, file : str = None ):
        super().__init__()
        
        self.reporter = ErrorReporter( "ParseError", str( err ).split('Expected')[ 0 ], contents, ( err.pos_in_stream, err.pos_in_stream + 1 ), file )

    def __repr__ ( self ):
        return str( self.reporter )

    def __str__ ( self ):
        return str( self.reporter )


from .keyboard import KeyboardLibrary, KeyStroke, PianoKey
from .keyboard_mido import KeyboardMidoLibrary, KeyboardMidoEventSource
from .keyboard_pynput import KeyboardPynputLibrary, KeyboardPynputEventSource
from .std import StandardLibrary
from .music import MusicLibrary
from .midi import MidiLibrary


from .transformer import Transformer
from musikla.core import Voice
from musikla.core.events import MusicEvent, NoteEvent, RestEvent, ChordEvent, DurationEvent
from typing import List, Optional, Tuple

class SlidingAverage():
    def __init__ ( self, capacity : int = 0 ):
        self.history : List[float] = []
        self.capacity = capacity
        self.sum : float = 0
        self.count : int = 0

    def __iadd__ ( self, value : float ):
        if self.count == 0:
            self.sum = value
            self.count = 1

            if self.capacity > 0:
                self.history.append( value )
        else:
            #  ( ( self.average / self.count ) + ( value / ( self.count + 1 ) ) )
            self.sum += value

            if self.capacity == 0 or self.count < self.capacity:
                self.count += 1
            else:
                self.sum -= self.history.pop( 0 )

            if self.capacity > 0:
                self.history.append( value )

        return self

    def __int__ ( self ):
        if self.count == 0:
            return 0

        return int( self.sum / self.count )
        
    def __float__ ( self ):
        if self.count == 0:
            return 0

        return float( self.sum / self.count )

class VoiceIdentifierVoice():
    def __init__ ( self, transformer : 'VoiceIdentifierTransformer', parent : Voice, index : int, staff : int = None ):
        self.transformer : VoiceIdentifierTransformer = transformer
        self.parent_name : str = parent.name
        self.parent_staff : Optional[int] = staff
        self.voice : Voice = parent.clone( parent.name + '/' + str( index ) )
        self.index = index
        self.average_pitch : SlidingAverage = SlidingAverage()
        self.auto_create_rests : bool = True
        self.auto_create_end_rests : bool = True
        self.last_event : Optional[MusicEvent] = None

    @property
    def last_timestamp ( self ):
        if self.last_event is None:
            return 0

        return self.last_event.timestamp
    
    @property
    def last_end_timestamp ( self ):
        if not self.last_event:
            return 0

        return self.last_event.end_timestamp
    
    @property
    def last_duration ( self ):
        if not self.last_event:
            return 0

        return self.last_event.duration

    def overlap_percentages ( self, timestamp : int, duration : int ) -> Tuple[float, float]:
        """
        This function returns two values, both between 0 and 1, that represent
        how much of each event is overlapping. For example, if the overlap is of
        50ms, and last event lasted 200ms, while the new event lasts 100ms.
        The result of this function would be (0.25, 0.5), which means 25% and 50%
        respectively.

        It should be called only when self.last_end_timestamp > timestamp
        """

        overlap_dur = self.last_end_timestamp - timestamp

        # When last_duration is == 0, we consider the overlapping percentage
        # to be 1, which means 100%
        overlap_perc1 = overlap_dur / self.last_duration if self.last_duration > 0 else 1

        overlap_perc2 = overlap_dur / duration if duration > 0 else 1

        return (overlap_perc1, overlap_perc2)

    def is_busy_at ( self, timestamp : int, duration : int = None ) -> bool:
        if self.last_end_timestamp <= timestamp:
            return False

        if duration is None:
            return True
        
        over1, over2 = self.overlap_percentages( timestamp, duration )

        # TODO Allow cutting emitted events
        # return over1 > self.transformer.autofix_overlaps_percentage \
        #    and over2 > self.transformer.autofix_overlaps_percentage
        return over2 > self.transformer.autofix_overlaps_percentage

    def is_connected_at ( self, timestamp : int ) -> bool:
        # TODO Implement overlap control here too
        return abs( timestamp - self.last_end_timestamp ) <= 3

    def distance ( self, event : NoteEvent ) -> float:
        return abs( int( event ) - int( self.average_pitch ) )

    def create_rest ( self, target : int, visible : bool = True ) -> RestEvent:
        duration = target - self.last_end_timestamp

        value = self.voice.from_duration_absolute( duration )

        return RestEvent(
            timestamp = self.last_end_timestamp,
            visible = visible,
            duration = duration,
            value = value,
            voice = self.voice,
            staff = self.parent_staff
        )

    def append ( self, *events : MusicEvent ):
        for event in events:
            if isinstance( event, NoteEvent ) or isinstance( event, ChordEvent ):
                self.average_pitch += int( event ) if isinstance( event, NoteEvent ) \
                                else  sum( event.pitches ) / len( event.pitches )

                # TODO Allow setting a minimum rest duration. If the empty space between the events is less than
                # said minimum, then no rest is created
                if self.auto_create_rests and self.last_end_timestamp < event.timestamp - 1:
                    self.transformer.add_output( self.create_rest( event.timestamp ) )

            if self.last_end_timestamp > event.timestamp:
                # o1, o2 = self.overlap_percentages( event.timestamp, event.duration )
                event = event.clone(
                    timestamp = self.last_end_timestamp,
                    end_timestamp = event.end_timestamp
                )

            self.transformer.add_output( event.clone( voice = self.voice ) )

            self.last_event = event

class VoiceIdentifierTransformer(Transformer):
    """
    This class receives a flat, ordered stream of musical events and splits them in multiple voices and identifies parallel notes/chords
    """
    def __init__ ( self, auto_create_rests : bool = True, auto_create_end_rests : bool = True ):
        super().__init__()

        self.voices : List[VoiceIdentifierVoice] = []

        self.auto_create_rests : bool = auto_create_rests
        
        self.auto_create_end_rests : bool = auto_create_end_rests

        # Value between 0 and 1
        self.autofix_overlaps_percentage : float = 0.1
    
    def create_voice_for ( self, event : DurationEvent ) -> VoiceIdentifierVoice:
        index = max( ( voice.index for voice in self.voices if voice.parent_name == event.voice.name ), default = 0 )

        voice = VoiceIdentifierVoice( self, event.voice, index + 1, event.staff )

        voice.auto_create_rests = self.auto_create_rests
        voice.auto_create_end_rests = self.auto_create_end_rests

        self.voices.append( voice )

        return voice

    def find_voice_for ( self, event : DurationEvent, auto_create : bool = True ) -> Optional[VoiceIdentifierVoice]:
        best_voice, best_voice_dst = None, None

        for voice in self.voices:
            if voice.parent_name != event.voice.name or voice.parent_staff != event.staff:
                continue

            # If the voice already has a sound playing at this timestamp
            if voice.is_busy_at( event.timestamp, event.duration ):
                continue

            if event.staff is not None:
                return voice

            voice_dst = voice.distance( event )

            if best_voice is None or best_voice_dst > voice_dst:
                best_voice = voice
                best_voice_dst = voice_dst

        if best_voice is None and auto_create:
            best_voice = self.create_voice_for( event )

        return best_voice
    
    def find_voice_for_rest ( self, event, auto_create : bool = True ) -> VoiceIdentifierVoice:
        for voice in self.voices:
            if voice.parent_name != event.voice.name or voice.parent_staff != event.staff:
                continue

            # If the voice already has a sound playing at this timestamp
            if voice.is_connected_at( event.timestamp ):
                return voice

        return self.create_voice_for( event )

    def register_rest ( self, event : RestEvent ):
        voice = self.find_voice_for_rest( event )

        if voice is not None:
            voice.append( event )

    def register_note ( self, event : NoteEvent ):
        voice = self.find_voice_for( event )

        if voice is not None:
            voice.append( event )
    
    def register_chord ( self, event : ChordEvent ):
        voice = self.find_voice_for( event )

        if voice is not None:
            voice.append( event )

    def transform ( self ):
        """
        For now, NoteOn/NoteOff events will be ignored. 
        Use the compose transformer before passing events into this notation builder. 
        This might change in the future.
        """
        while True:
            done, event = yield

            if done: break

            if isinstance( event, NoteEvent ):
                self.register_note( event )
            elif isinstance( event, RestEvent ):
                self.register_rest( event )
            elif isinstance( event, ChordEvent ):
                self.register_chord( event )
            else:
                self.add_output( event )

        if self.auto_create_end_rests and self.voices:
            last_voice_timestamp : int = max( [ voice.last_end_timestamp for voice in self.voices ] )

            for voice in self.voices:
                if voice.last_end_timestamp < last_voice_timestamp:
                    self.add_output( voice.create_rest( last_voice_timestamp, visible = False ) )



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .context_modifier_node import ContextModifierNode
from musikla.core.events import ContextChangeEvent
from musikla.core import Context, Voice

class OctaveModifierNode( ContextModifierNode ):
    def __init__ ( self, octave, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.octave = octave
        
    def apply ( self, voice : Voice ):
        voice.octave = self.octave

    def modify ( self, context : Context ):
        context.voice = context.voice.clone( octave = self.octave )

        yield ContextChangeEvent( context.cursor, "octave", context.voice.octave, context.voice, 0 )
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'O' + str( self.octave ) )


from typing import Tuple, Optional
from musikla.parser.printer import CodePrinter
from musikla.core import Context, Music
from ..node import Node

class GroupNode( Node ):
    def __init__ ( self, expression : Node = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )
        
        self.expression : Optional[Node] = expression

    def get_events ( self, context : Context, forked : Context, value : Music ):
        try:
            for event in value.expand( context ):
                yield event
        finally:
            context.join( forked )

    def __eval__ ( self, context : Context, assignment : bool = False ):
        if self.expression is None:
            return None

        forked = context.fork()

        value = self.expression.eval( forked )

        if isinstance( value, Music ):
            return Music( self.get_events( context, forked, value ) )
        else:
            context.join( forked )

            return value

    def to_source ( self, printer : CodePrinter ):
        with printer.block( '(', ')' ):
            if self.expression is not None:
                self.expression.to_source( printer )




from sys import exc_info
from musikla.parser.printer import CodePrinter
from typing import Any, Optional, Tuple
from musikla.core import Context 
from ..error_reporter import ErrorReporter

class Node():
    def __init__ ( self, position : Tuple[int, int, int] = None ):
        self.position : Optional[Tuple[int, int, int]] = position
        self.hoisted : bool = False

    def __eval__ ( self, context : Context ) -> Any:
        return None
    
    def eval ( self, context : Context ) -> Any:
        try:
            return self.__eval__( context )
        except:
            _, err, _ = exc_info()

            if err is not None:
                if isinstance( err, MusiklaExecutionError ):
                    if err.position is None:
                        err.position = self.position
                    
                    raise err from None
                else:
                    raise MusiklaExecutionError( err, self.position )

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( f"<{self.__class__.__name__}>" )
        
    def __repr__ ( self ):
        return "<%s>(%r)" % (self.__class__.__name__, self.__dict__)

class ValueNode(Node):
    def __init__ ( self, value : Any ):
        self.value : Any = value
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( "<VALUE>" )

    def __eval__ ( self, context : Context ):
        return self.value

class MusiklaExecutionError(Exception):
    def __init__ ( self, base_exception : BaseException, position : Tuple[int, int, int] = None ):
        super().__init__()

        self.position : Optional[Tuple[int, int, int]] = position
        self.base_exception : BaseException = base_exception
        self.__traceback__ = base_exception.__traceback__
        self.file : Optional[str] = None
        self.content : Optional[str] = None
        self.reporter : Optional[ErrorReporter] = None
    
    def create_reporter ( self, file : str = None, content : str = None ):
        self.content = content
        self.file = file

        if self.position is not None and self.content is not None and self.file is not None:
            _, start, end = self.position
            
            self.reporter = ErrorReporter( "EvalError", str( self.base_exception ), self.content, ( start, end ), self.file )
    
    def __str__ ( self ):
        if self.reporter is not None:
            return str( self.reporter )
        
        return str( self.base_exception )



from typing import Any, Iterator, Optional, Tuple
from .node import Node
from .expressions import ExpressionNode
from musikla.core import Music, StackFrame


class MusicSequenceBase( Node ):
    def check_returns ( self ):
        return True

    def values ( self, context ):
        return iter(())

    def _check_stack_frame ( self, stack_frame : Optional[StackFrame], music_mode : bool = True ) -> bool:
        if not self.check_returns():
            return False

        if stack_frame is not None and stack_frame.returned:
            if music_mode and stack_frame.returned_value is not None:
                raise Exception( "A function that plays music cannot return value" )
            else:
                return True
        
        return False

    def _get_events ( self, context, stack_frame : Optional[StackFrame], iterator : Iterator[Any], pending_value = None ):
        if isinstance( pending_value, Music ):
            for event in pending_value.expand( context ):
                yield event

        if self._check_stack_frame( stack_frame ):
            return

        for value in iterator:
            if isinstance( value, Music ):
                for event in value.expand( context ):
                    yield event

            if self._check_stack_frame( stack_frame ):
                break

    def __eval__ ( self, context ):
        value = None

        stack_frame : Optional[StackFrame] = context.symbols.lookup( 'stack_frame', container = 'stack' )

        iterator = iter( self.values( context ) )

        while True:
            try:
                value = next( iterator )

                if self._check_stack_frame( stack_frame, False ):
                    return stack_frame.returned_value

                if isinstance( value, Music ):
                    return Music( self._get_events( context, stack_frame, iterator, value ) )
            except StopIteration:
                break

        return value


class MusicNode( ExpressionNode ):
    def __init__ ( self, position : Tuple[int, int, int] = None ):
        super().__init__( position )
    
    def get_events ( self, context ):
        return iter( () )

    def __eval__ ( self, context ):
        return Music( self.get_events( context ) )



from typing import Any
from ..node import Node
from musikla.core import Context

class ExpressionNode( Node ):
    def __eval__ ( self, context : Context ) -> Any:
        raise BaseException("Abstract Expression cannot be evaluated")



from chunk import Chunk
from typing import Any, Generator, cast, IO
from itertools import islice
import struct
import io

class SoundFont:
    @staticmethod
    def from_file ( file : str ) -> 'SoundFont':
        return SoundFont( cast( IO[bytes], open( file, 'rb' ) ) )

    @staticmethod
    def from_io ( stream : IO[bytes] ) -> 'SoundFont':
        return SoundFont( stream )

    def __init__ ( self, stream : IO[bytes] ):
        self.stream : IO[bytes] = stream

        for ch in islice( self.read_chunks(), 10 ):
            name = ch.chunkname.decode( 'utf-8', errors = 'replace' )

            print( name, ch.chunksize )

            if ( hasattr( self, 'parse_' + name ) ):
                getattr( self, 'parse_' + name )( ch )
    
    def parse_RIFF ( self, chunk : Chunk ):
        print( chunk.read( 4 ).decode( 'utf-8', errors = 'replace' ) )
        pass
    
    def parse_LIST ( self, chunk : Chunk ):
        print( chunk.read( 4 ).decode( 'utf-8', errors = 'replace' ) )
        pass

    def parse_ifil ( self, chunk : Chunk ):
        chunk.seek( chunk.tell() + 4 )
        print( chunk.read( 4 ).decode( 'utf-8', errors = 'replace' ) )
        pass

    def read_chunks ( self ) -> Generator[Chunk, Any, Any]:
        try:
            while True:
                print( self.stream.tell() )

                ch : Chunk = Chunk( self.stream, True )

                yield ch
        except EOFError:
            pass

sf2 = SoundFont.from_file( 'C:\\Users\\PedroSilva\\Downloads\\freepiano_2.2.2.1_win64\\freepiano\\sf2\\FluidR3_GM.sf2' )



from musikla.core.events import sound
from musikla.core.voice import Voice
from musikla.core.instrument import Instrument
import pyfluidsynth
from musikla.core import Clock
from musikla.core.events import MusicEvent, VoiceEvent, NoteEvent, NoteOnEvent, NoteOffEvent, SoundEvent, ProgramChangeEvent, ControlChangeEvent, CallbackEvent
from musikla.core.events.transformers import DecomposeChordsTransformer
from .sequencer import Sequencer, SequencerFactory, ArgumentParser
from ctypes import c_void_p, c_int
from threading import Semaphore
from typing import Dict, List, Mapping, Tuple, Any, Optional, Union
from time import sleep
from collections import defaultdict
import re

INT_PATTERN = re.compile("^([0-9]+)$")
FLOAT_PATTERN = re.compile("^([0-9]*\.[0-9]+)$")

fluid_synth_get_active_voice_count = pyfluidsynth.cfunc('fluid_synth_get_active_voice_count', c_int,
                                    ('synth', c_void_p, 1))

class FluidSynthSequencer ( Sequencer ):
    def __init__ ( self, output : str = None, soundfonts : Dict[str, Optional[str]] = {}, settings : Mapping[str, Any] = {}, realtime : bool = True, device : str = None ):
        super().__init__()

        self.realtime = realtime

        self.output : str = output or "pulseaudio"
        self.device : Optional[str] = device
        self.settings : Mapping[str, Any] = settings

        self.synth : Optional[pyfluidsynth.Synth] = None
        self.synthId : Optional[int] = None
        self.soundfonts : Dict[str, int] = dict( [ ( sf, -1 ) for sf in soundfonts.keys() ] )
        self.soundfont_aliases : Dict[str, str] = dict( [ ( alias, sf ) for sf, alias in soundfonts.items() if alias is not None ] )
        self.defaultSoundfont : Optional[int] = None
        self.ramSoundfont : Optional[pyfluidsynth.RamSoundFont] = None
        self.ramSoundfontId : Optional[int] = None
        self.sequencer : Optional[pyfluidsynth.Sequencer] = None
        self.client : Optional[int] = None

        self.last_note : Optional[int] = None
        self.join_lock : Optional[Semaphore] = None
        self.start_time : int = 0
        self.keys_count : Dict[int, List[int]] = defaultdict(lambda: [ 0 for i in range( 0, 128 ) ])

        self.events_data : Dict[int, Any] = dict()
        self.events_data_id : int = 1
        self.voices : Dict[str, int] = dict()
        self.samples : Dict[int, Tuple[int, int]] = dict()
        self.samples_voices : Dict[int, Voice] = dict()
        self.last_sample : int = 0
        # For some reason, notes below 15 are somewhat pitched down, and below 10 are heavily pitched down
        # So all samples must start above 15. This means that instead of 128 notes available, we have 113
        self.sample_note_offset : int = 15

        self.clock : Clock = Clock()

        self.set_transformers(
            DecomposeChordsTransformer()
        )
    
    def _get_voice_channel ( self, voice : Voice ) -> int:
        key = voice.name + '$' + str( voice.instrument.program )

        if key in self.voices:
            return self.voices[ key ]

        value = len( self.voices )

        self.voices[ key ] = value

        if voice.instrument.soundfont is None and voice.instrument.bank is None:
            self.synth.program_change( value, voice.instrument.program - 1 )
        elif voice.instrument.bank is not None:
            soundfont = self.get_soundfont_id( voice.instrument.soundfont )

            self.synth.program_select( value, soundfont, voice.instrument.bank, voice.instrument.program - 1 )
        
        return value

    def _sound_to_note_event ( self, event : SoundEvent ) -> NoteEvent:
        if event.file not in self.samples:
            sno : int = self.sample_note_offset

            sample_location = ( self.last_sample // ( 128 - sno ) + 1, sno + self.last_sample % ( 128 - sno ) )

            self.ramSoundfont.add_wave_zone( 1, sample_location[ 0 ], event.wave, sample_location[ 1 ], sample_location[ 1 ] )
            
            self.samples[ event.file ] = sample_location

            self.last_sample += 1
        else:
            sample_location = self.samples[ event.file ]

        if sample_location[ 0 ] not in self.samples_voices:
            voice = event.voice.clone( name = f"$$sample_voice_{ len( self.samples_voices ) }" )

            voice.instrument = Instrument( f"Samples Virtual Instrument { len( self.samples_voices ) }", sample_location[ 0 ], 1, self.ramSoundfontId )

            self.samples_voices[ sample_location[ 0 ] ] = voice
        else:
            voice = self.samples_voices[ sample_location[ 0 ] ]

        return NoteEvent.from_pitch(
            timestamp = event.timestamp,
            duration = event.duration,
            value = event.value,
            pitch = sample_location[ 1 ],
            velocity = event.velocity,
            voice = voice
        )

    @property
    def is_output_file ( self ) -> bool:
        if self.output != None and isinstance( self.output, str ):
            return '.' in self.output
        else:
            return False

    @property
    def playing ( self ):
        if self.synth is None:
            return False

        now = self.get_time()

        return self.last_note != None and now <= self.last_note
    
    @property
    def started ( self ):
        return self.synth is not None

    def load_soundfont ( self, soundfont : str, alias : str = None ) -> int:
        if self.started:
            sf_id = self.synth.sfload( soundfont, update_midi_preset = 1 )

            if self.defaultSoundfont is None:
                self.defaultSoundfont = sf_id
        else:
            sf_id = -1

        self.soundfonts[ soundfont ] = sf_id

        if alias is not None:
            self.soundfont_aliases[ alias ] = soundfont

        return sf_id

    def unload_soundfont ( self, soundfont : str ):
        if soundfont in self.soundfont_aliases:
            alias, soundfont = soundfont, self.soundfont_aliases[ soundfont ]

            del self.soundfont_aliases[ alias ]

        if self.started:
            self.synth.sfunload( self.soundfonts[ soundfont ], update_midi_preset = 1 )

        if soundfont in self.soundfonts:
            del self.soundfonts[ soundfont ]

    def get_soundfont_id ( self, soundfont : Union[str, int, None] ) -> Optional[int]:
        if type( soundfont ) is int:
            return soundfont
        elif type( soundfont ) is str:
            if soundfont in self.soundfont_aliases:
                soundfont = self.soundfont_aliases[ soundfont ]
            
            if soundfont in self.soundfonts:
                return self.soundfonts[ soundfont ]
        elif soundfont is None:
            return self.defaultSoundfont
        
    def get_time ( self ):
        if self.sequencer == None:
            return 0
        
        return self.sequencer.get_tick() - self.start_time

    def timer ( self, time, data = None, source = -1, dest = -1, absolute = True ):
        if self.last_note == None or self.last_note < time:
            self.last_note = time

        if data != None:
            self.events_data[ self.events_data_id ] = data

            data = c_void_p( self.events_data_id )

            self.events_data_id += 1

        # print( time, data, source, dest, absolute )
        return self.sequencer.timer( time, data = data, source = source, dest = dest, absolute = absolute )

    def apply_event_callback ( self, time, event, seq, data ):
        data = pyfluidsynth.fluid_event_get_data( event )

        if data != None:
            data_id = data

            data = self.events_data[ data_id ]

            del self.events_data[ data_id ]

        if data != None:
            self.apply_event( data )
        else:
            # TODO HACK!!!
            self._join_callback( time, event, seq, data )

    def apply_event_noteoff ( self, channel, event ):
        key = int( event )

        self.keys_count[ channel ][ key ] -= 1

        if self.keys_count[ channel ][ key ] == 0:
            self.synth.noteoff( channel, key )

    def apply_event_noteon( self, channel, event ):
        key = int( event )

        self.keys_count[ channel ][ key ] += 1

        self.synth.noteon( channel, key, event.velocity )
    
    def apply_channel_program ( self, channel : int, soundfont : int, bank : int, program : int ):
        # TODO Cache current program to avoid continuous calls to fluidsynth
        self.synth.program_select( channel, soundfont, bank, program )

    def apply_event ( self, event : MusicEvent ):
        if isinstance( event, VoiceEvent ):
            channel = self._get_voice_channel( event.voice )

            soundfont = self.get_soundfont_id( event.voice.instrument.soundfont )

            if soundfont is not None:
                self.apply_channel_program( channel, soundfont, event.voice.instrument.bank or 0, event.voice.instrument.program )
                        
            if isinstance( event, NoteOnEvent ):
                self.apply_event_noteon( channel, event )
            elif isinstance( event, NoteOffEvent ):
                self.apply_event_noteoff( channel, event )
            elif isinstance( event, ProgramChangeEvent ):
                self.synth.program_change( channel, event.program )
            elif isinstance( event, ControlChangeEvent ):
                self.synth.cc( channel, event.control, event.value )
        elif isinstance( event, CallbackEvent ):
            event.call()
        else:
            pass

    def on_event ( self, event : MusicEvent ):
        if isinstance( event, VoiceEvent ):
            self._get_voice_channel( event.voice )
        
        if isinstance( event, SoundEvent ):
            event = self._sound_to_note_event( event )
            
        if isinstance( event, NoteEvent ):
            noteon = event.note_on
            noteoff = event.note_off
            
            self.timer( self.start_time + noteon.timestamp, data = noteon, dest = self.client )
            self.timer( self.start_time + noteoff.timestamp, data = noteoff, dest = self.client )    
        else:
            self.timer( event.timestamp, data = event, dest = self.client )
    
    def on_close ( self ):
        # Render the file
        if not self.realtime:
            pyfluidsynth.FileRenderer.fast_loop_sequencer( self.synth, self.sequencer, self.last_note + 1000 )

        # Release resources
        self.sequencer.delete()
        self.synth.delete()

    def _join_callback ( self, time, event, seq, data ):
        if self.join_lock is not None:
            self.join_lock.release()

            self.join_lock = None

    def join ( self ):
        if not self.playing:
            return

        if self.realtime:
            self.join_lock = Semaphore( 0 )

            self.timer( self.last_note + 1000, dest = self.client )
            
            self.join_lock.acquire()

        # while True:
        #     voices = fluid_synth_get_active_voice_count( self.synth.synth )

        #     print( voices )

        #     sleep( 0.3 )

        #     if voices == 0:
        #         break

    def start ( self ):
        if self.started:
            return

        self.synth = pyfluidsynth.Synth( **self.settings )

        if self.is_output_file:
            self.synth.setting( 'audio.file.name', self.output )
            self.synth.setting( 'audio.driver', 'file' )
            
            if self.realtime:
                self.synth.audio_driver = pyfluidsynth.new_fluid_audio_driver( self.synth.settings, self.synth.synth )
        else:
            if not self.realtime:
                raise BaseException( "Can only render audio fast to a file (not audio device)" )

            self.synth.start( driver = self.output, device = self.device )

        for soundfont in self.soundfonts.keys():
            self.load_soundfont( soundfont )

        self.ramSoundfont = pyfluidsynth.RamSoundFont()
        self.ramSoundfontId = self.synth.add_sfont( self.ramSoundfont )
        
        self.synth.cc( 0, 64, 127 )
    
        self.sequencer = pyfluidsynth.Sequencer( use_system_timer = self.realtime )
        
        self.synthId = self.sequencer.register_fluidsynth( self.synth )

        self.client = self.sequencer.register_client( "eventClient", self.apply_event_callback )

        self.start_time = self.sequencer.get_tick()

class FluidSynthSequencerFactory( SequencerFactory ):
    default : bool = True

    def init ( self ):
        self.name = 'fluidsynth'
        self.argparser = ArgumentParser( description = 'Synthesize notes throught the FluidSynthesizer library' )
        self.argparser.add_argument( '-d', '--device', dest = 'device', type = str, action='store', help = 'Specify a custom device name for the sound output' )
        self.argparser.add_argument( '-s', '--setting', dest = 'settings', type = str, action='append', help = 'Pass a setting down to fluidsynth with the format <key>=<value>' )
        self.argparser.add_argument( '-c', '--audio-bufcount', dest = 'audio_buffcount', type = int, action='store', help = 'Number of audio buffers (equivalent to `audio.periods`)' )
        self.argparser.add_argument( '-g', '--gain', dest = 'gain', type = float, action='store', help = 'Set the master gain [0 < gain < 10, default = 0.2] (equivalent to `synth.gain`)' )
        self.argparser.add_argument( '-K', '--midi-channels', dest = 'midi_channels', type = int, action='store', help = 'The number of midi channels [default = 16] (equivalent to `synth.midi-channels`)' )
        self.argparser.add_argument( '-L', '--audio-channels', dest = 'audio_channels', type = int, action='store', help = 'The number of stereo audio channels [default = 1] (equivalent to `synth.audio-channels`)' )
        self.argparser.add_argument( '-r', '--sample-rate', dest = 'sample_rate', type = float, action='store', help = 'Set the sample rate (equivalent to `synth.sample-rate`)' )
        self.argparser.add_argument( '-z', '--audio-bufsize', dest = 'audio_buffsize', type = float, action='store', help = 'Size of each audio buffer (equivalent to `audio.period-size`)' )
        self.argparser.add_argument( '--fast', dest = 'fast', action='store_true', help = 'Render the music file as fast as possible (instead of in realtime)' )

        self.soundfonts : Dict[str, Optional[str]] = {}

    def _args_settings_dictionary ( self, args ):
        settings_dict = {}

        for arg in args.settings or []:
            if '=' not in arg:
                settings_dict[ arg ] = True
            else:
                key, value = arg.split( '=' )

                if value == 'true':
                    settings_dict[ key ] = True
                elif value == 'false':
                    settings_dict[ key ] = False
                elif FLOAT_PATTERN.match( value ):
                    settings_dict[ key ] = float( value )
                elif INT_PATTERN.match( value ):
                    settings_dict[ key ] = int( value )
                else:
                    settings_dict[ key ] = value

        if args.audio_buffcount is not None:
            settings_dict[ 'audio.periods' ] = args.audio_buffcount
        if args.gain is not None:
            settings_dict[ 'synth.gain' ] = args.gain
        if args.midi_channels is not None:
            settings_dict[ 'synth.midi-channels' ] = args.midi_channels
        if args.audio_channels is not None:
            settings_dict[ 'synth.audio-channels' ] = args.audio_channels
        if args.sample_rate is not None:
            settings_dict[ 'synth.sample-rate' ] = args.sample_rate
        if args.audio_buffsize is not None:
            settings_dict[ 'audio.period-size' ] = args.audio_buffsize

        return settings_dict

    def add_soundfont ( self, soundfont : str, alias : str = None ):
        self.soundfonts[ soundfont ] = alias

    def remove_soundfont ( self, soundfont : str ):
        keys = []
        for key, alias in self.soundfonts.keys():
            if alias == soundfont:
                keys.append( key )
        
        for key in keys:
            del self.soundfonts[ key ]
        
        if soundfont in self.soundfonts:
            del self.soundfonts[ soundfont ]

    def from_str ( self, uri : str, args ) -> FluidSynthSequencer:
        soundfont = self.config.get( 'Musikla', 'soundfont', fallback = "/usr/share/sounds/sf2/FluidR3_GM.sf2" )

        cli_settings = self._args_settings_dictionary( args )
        ini_settings = self.config[ 'FluidSynth.Settings' ] if 'FluidSynth.Settings' in self.config else {}

        soundfonts = { **self.soundfonts }

        if soundfont is not None:
            soundfonts[ soundfont ] = None

        return FluidSynthSequencer( uri, soundfonts, { **cli_settings, **ini_settings }, realtime = not args.fast, device = args.device )
